@@ -72,7 +72,8 @@
     delete/1,
     reset_counters/1,
     reset_counters/2,
-    start/0
+    start/0,
+    terminate/2
 ]).
 
 -export([
@@ -203,22 +204,32 @@ union(Queries) ->
 %% The counters are reset by default, unless Reset is set to false 
 -spec compile(atom(), glc_ops:op() | [glc_ops:op()]) -> {ok, atom()}.
 compile(Module, Query) ->
-    compile(Module, Query, undefined, true).
+    compile(Module, Query, [{statistics, true}]).
 
--spec compile(atom(), glc_ops:op() | [glc_ops:op()], boolean()) -> {ok, atom()}.
+-spec compile(atom(), glc_ops:op() | [glc_ops:op()], atom() | list() | boolean()) -> {ok, atom()}.
+compile(Module, Query, Store) when not is_boolean(Store) ->
+    compile(Module, Query, Store, true);
 compile(Module, Query, Reset) when is_boolean(Reset) ->
-    compile(Module, Query, undefined, Reset);
-compile(Module, Query, undefined) ->
-    compile(Module, Query, undefined, true);
-compile(Module, Query, Store) when is_list(Store) ->
-    compile(Module, Query, Store, true).
+    compile(Module, Query, undefined, Reset).
 
-compile(Module, Query, Store, Reset) ->
-    {ok, ModuleData} = module_data(Module, Query, Store),
-    case glc_code:compile(Module, ModuleData) of
-        {ok, Module} when Reset ->
+compile(Module, Query, Store, Reset) when Store =:= []; Store =:= undefined ->
+    compile(Module, Query, [{statistics, true}], Reset);
+compile(Module, Query, Store, Reset) when is_list(Store) ->
+    case lists:keyfind(statistics, 1, Store) of
+        {_, true} -> 
+            compile(Module, Query, Store, true, Reset);
+        _ -> 
+            compile(Module, Query, Store, false, false)
+    end.
+
+compile(Module, Query, Store, Stats, Reset) ->
+    {ok, ModuleData} = module_data(Module, Query, Store, Stats),
+    case glc_code:compile(Module, ModuleData, Stats) of
+        {ok, Module} when Stats =:= true, Reset =:= true ->
             reset_counters(Module),
             {ok, Module};
+        {ok, Module} when Stats =:= true ->
+            {ok, Module};
         {ok, Module} ->
             {ok, Module}
     end.
@@ -284,6 +295,36 @@ job_input(Module) ->
 job_time(Module) ->
     Module:info(job_time).
 
+%% @doc Terminate a modules supervisors
+-spec terminate(atom(), all | counters) -> ok.
+terminate(Module, counters) ->
+    Counts = counts_name(Module),
+    ManageCounts = manage_counts_name(Module),
+
+    _ = [ begin 
+        ok = supervisor:terminate_child(Sup, Name),
+        ok = supervisor:delete_child(Sup, Name)
+      end || {Sup, Name} <- 
+        [{gr_manager_sup, ManageCounts},
+         {gr_counter_sup, Counts}]
+    ],
+    ok;
+terminate(Module, params) ->
+    Params = params_name(Module),
+    ManageParams = manage_params_name(Module),
+
+    _ = [ begin 
+        ok = supervisor:terminate_child(Sup, Name),
+        ok = supervisor:delete_child(Sup, Name)
+      end || {Sup, Name} <- 
+        [{gr_manager_sup, ManageParams},
+         {gr_param_sup, Params}]
+    ],
+    ok;
+terminate(Module, all) ->
+    catch (terminate(Module, counters)), % Catch on no statistics option
+    terminate(Module, params).
+
 %% @doc Release a compiled query.
 %%
 %% This releases all resources allocated by a compiled query. The query name
@@ -291,19 +332,7 @@ job_time(Module) ->
 %% function will shutdown all relevant processes and purge/delete the module.
 -spec delete(atom()) -> ok.
 delete(Module) ->
-    Params = params_name(Module),
-    Counts = counts_name(Module),
-    ManageParams = manage_params_name(Module),
-    ManageCounts = manage_counts_name(Module),
-
-    _ = [ begin 
-        ok = supervisor:terminate_child(Sup, Name),
-        ok = supervisor:delete_child(Sup, Name)
-      end || {Sup, Name} <- 
-        [{gr_manager_sup, ManageParams}, {gr_manager_sup, ManageCounts},
-         {gr_param_sup, Params}, {gr_counter_sup, Counts}]
-    ],
-
+    ok = terminate(Module, all),
     code:soft_purge(Module),
     code:delete(Module),
     ok.
@@ -322,9 +351,16 @@ reset_counters(Module) ->
 reset_counters(Module, Counter) ->
     Module:reset_counters(Counter).
 
+prepare_store(Store) when not is_list(Store) -> Store;
+prepare_store(Store) ->
+    lists:map(fun({K, V}) when is_pid(V); is_port(V); is_reference(V) 
+                        -> {K, {other, binary_to_list(term_to_binary(V))}} ;
+                 ({K, V}) -> {K, V} 
+          end, Store).
+
 %% @private Map a query to a module data term.
--spec module_data(atom(), term(), term()) -> {ok, #module{}}.
-module_data(Module, Query, Store) ->
+-spec module_data(atom(), term(), term(), boolean()) -> {ok, #module{}}.
+module_data(Module, Query, Store, Stats) ->
     %% terms in the query which are not valid arguments to the
     %% erl_syntax:abstract/1 functions are stored in ETS.
     %% the terms are only looked up once they are necessary to
@@ -335,33 +371,43 @@ module_data(Module, Query, Store) ->
     %% the abstract_tables/1 function expects a list of name-atom pairs.
     %% tables are referred to by name in the generated code. the table/1
     %% function maps names to registered processes response for those tables.
-    Tables = module_tables(Module),
+    Tables = module_tables(Module, Stats),
     Query2 = glc_lib:reduce(Query),
-    {ok, #module{'query'=Query, tables=Tables, qtree=Query2, store=Store}}.
+    Store2 = prepare_store(Store),
+    {ok, #module{'query'=Query, tables=Tables, qtree=Query2, store=Store2}}.
 
 %% @private Create a data managed supervised process for params, counter tables
-module_tables(Module) ->
+-spec module_tables(atom(), boolean()) -> list().
+module_tables(Module, Stats) ->
     Params = params_name(Module),
     Counts = counts_name(Module),
     ManageParams = manage_params_name(Module),
     ManageCounts = manage_counts_name(Module),
-    Counters = [{input,0}, {filter,0}, {output,0}, 
-                {job_input, 0}, {job_run,0},  {job_time, 0},
-                {job_error, 0}],
 
     _ = supervisor:start_child(gr_param_sup, 
         {Params, {gr_param, start_link, [Params]}, 
         transient, brutal_kill, worker, [Params]}),
-    _ = supervisor:start_child(gr_counter_sup, 
-        {Counts, {gr_counter, start_link, [Counts]}, 
-        transient, brutal_kill, worker, [Counts]}),
     _ = supervisor:start_child(gr_manager_sup, 
         {ManageParams, {gr_manager, start_link, [ManageParams, Params, []]},
         transient, brutal_kill, worker, [ManageParams]}),
-    _ = supervisor:start_child(gr_manager_sup, {ManageCounts, 
-        {gr_manager, start_link, [ManageCounts, Counts, Counters]},
-        transient, brutal_kill, worker, [ManageCounts]}),
-    [{params,Params}, {counters, Counts}].
+
+    Tables = case Stats of
+        true ->
+            Counters = [{input,0}, {filter,0}, {output,0}, 
+                        {job_input, 0}, {job_run,   0},  
+                        {job_time,  0}, {job_error, 0}],
+            _ = supervisor:start_child(gr_counter_sup, 
+                {Counts, {gr_counter, start_link, [Counts]}, 
+                transient, brutal_kill, worker, [Counts]}),
+            _ = supervisor:start_child(gr_manager_sup, {ManageCounts, 
+                {gr_manager, start_link, [ManageCounts, Counts, Counters]},
+                transient, brutal_kill, worker, [ManageCounts]}),
+            [{counters, Counts}];
+        false ->
+            [{counters, undefined}]
+     end,
+    [{params, Params} | Tables].
+
 
 reg_name(Module, Name) ->
     list_to_atom("gr_" ++ atom_to_list(Module) ++ Name).
@@ -406,11 +452,17 @@ start() ->
 -include_lib("eunit/include/eunit.hrl").
 
 setup_query(Module, Query) ->
-    setup_query(Module, Query, undefined).
+    setup_query(Module, Query, [{statistics, true}]).
 
 setup_query(Module, Query, Store) ->
-    ?assertNot(erlang:module_loaded(Module)),
-    ?assertEqual({ok, Module}, case (catch compile(Module, Query, Store)) of
+    setup_query(Module, Query, Store, true).
+
+setup_query(Module, Query, Store, Reset) ->
+    case Reset of
+        true  -> ?assertNot(erlang:module_loaded(Module));
+        false -> ?assert(erlang:module_loaded(Module))
+    end,
+    ?assertEqual({ok, Module}, case (catch compile(Module, Query, Store, Reset)) of
         {'EXIT',_}=Error -> ?debugFmt("~p", [Error]), Error; Else -> Else end),
     ?assert(erlang:function_exported(Module, table, 1)),
     ?assert(erlang:function_exported(Module, handle, 1)),
@@ -452,9 +504,18 @@ events_test_() ->
                     ?assertEqual({null, false}, Mod:info('query'))
                 end
             },
+            {"no init counters test",
+                fun() ->
+                    {compiled, Mod} = setup_query(testmod4a, glc:null(false), [{statistics, false}]),
+                    glc:handle(Mod, gre:make([], [list])),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output))
+                end
+            },
             {"init counters test",
                 fun() ->
-                    {compiled, Mod} = setup_query(testmod4, glc:null(false)),
+                    {compiled, Mod} = setup_query(testmod4b, glc:null(false)),
                     ?assertEqual(0, Mod:info(input)),
                     ?assertEqual(0, Mod:info(filter)),
                     ?assertEqual(0, Mod:info(output))
@@ -631,7 +692,7 @@ events_test_() ->
             {"with function storage test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, true}],
                     {compiled, Mod} = setup_query(testmod15,
                         glc:with(glc:eq(a, 1), fun(Event, EStore) -> 
                            Self ! {gre:fetch(a, Event), EStore} end),
@@ -643,7 +704,7 @@ events_test_() ->
             },
             {"delete test",
                 fun() ->
-                    {compiled, Mod} = setup_query(testmod16, glc:null(false)),
+                    {compiled, Mod} = setup_query(testmod16a, glc:null(false)),
                     ?assert(is_atom(Mod:table(params))),
                     ?assertMatch([_|_], gr_param:info(Mod:table(params))),
                     ?assert(is_list(code:which(Mod))),
@@ -661,9 +722,30 @@ events_test_() ->
                     ?assertEqual(undefined, whereis(manage_counts_name(Mod)))
                 end
             },
+            {"delete test no stats",
+                fun() ->
+                    {compiled, Mod} = setup_query(testmod16b, glc:null(false),
+                        [{statistics, false}]),
+                    ?assert(is_atom(Mod:table(params))),
+                    ?assertMatch([_|_], gr_param:info(Mod:table(params))),
+                    ?assert(is_list(code:which(Mod))),
+                    ?assert(is_pid(whereis(params_name(Mod)))),
+                    ?assertNot(is_pid(whereis(counts_name(Mod)))),
+                    ?assert(is_pid(whereis(manage_params_name(Mod)))),
+                    ?assertNot(is_pid(whereis(manage_counts_name(Mod)))),
+
+                    glc:delete(Mod),
+                    
+                    ?assertEqual(non_existing, code:which(Mod)),
+                    ?assertEqual(undefined, whereis(params_name(Mod))),
+                    ?assertEqual(undefined, whereis(counts_name(Mod))),
+                    ?assertEqual(undefined, whereis(manage_params_name(Mod))),
+                    ?assertEqual(undefined, whereis(manage_counts_name(Mod)))
+                end
+            },
             {"reset counters test",
                 fun() ->
-                    {compiled, Mod} = setup_query(testmod17,
+                    {compiled, Mod} = setup_query(testmod17a,
                         glc:any([glc:eq(a, 1), glc:eq(b, 2)])),
                     glc:handle(Mod, gre:make([{'a', 2}], [list])),
                     glc:handle(Mod, gre:make([{'b', 1}], [list])),
@@ -689,6 +771,80 @@ events_test_() ->
                     ?assertEqual(0, Mod:info(output))
                 end
             },
+            {"reset all counters test",
+                fun() ->
+                    {compiled, Mod} = setup_query(testmod17b,
+                        glc:any([glc:eq(a, 1), glc:eq(b, 2)])),
+                    glc:handle(Mod, gre:make([{'a', 2}], [list])),
+                    glc:handle(Mod, gre:make([{'b', 1}], [list])),
+                    ?assertEqual(2, Mod:info(input)),
+                    ?assertEqual(2, Mod:info(filter)),
+                    glc:handle(Mod, gre:make([{'a', 1}], [list])),
+                    glc:handle(Mod, gre:make([{'b', 2}], [list])),
+                    ?assertEqual(4, Mod:info(input)),
+                    ?assertEqual(2, Mod:info(filter)),
+                    ?assertEqual(2, Mod:info(output)),
+
+                    Self = self(),
+                    glc:run(Mod, fun(Event, EStore) -> 
+                        Self ! {gre:fetch(a, Event), EStore}
+                    end, [{a,1}]),
+
+                    glc:run(Mod, fun(Event, _EStore) -> 
+                        erlang:error(pow, Event)
+                    end, [{a,2}]),
+
+                    ?assertEqual(3, Mod:info(output)),
+                    ?assertEqual(3, Mod:info(filter)),
+                    ?assertEqual({1, [{statistics, true}]}, 
+                                 receive MsgStore -> 
+                                         MsgStore after 0 -> notcalled end),
+                    ?assertEqual(2, Mod:info(job_input)),
+                    ?assertEqual(1, Mod:info(job_run)),
+                    ?assert(0 < Mod:info(job_time)),
+                    ?assertEqual(1, Mod:info(job_error)),
+
+                    glc:reset_counters(Mod, all),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_run)),
+                    ?assertEqual(0, Mod:info(job_time)),
+                    ?assertEqual(0, Mod:info(job_error))
+                end
+            },
+            {"recompile without reset counters test",
+                fun() ->
+                    {compiled, Mod} = setup_query(testmod17c,
+                        glc:any([glc:eq(a, 1), glc:eq(b, 2)]), []),
+                    glc:handle(Mod, gre:make([{'a', 2}], [list])),
+                    glc:handle(Mod, gre:make([{'b', 1}], [list])),
+                    ?assertEqual(2, Mod:info(input)),
+                    ?assertEqual(2, Mod:info(filter)),
+                    glc:handle(Mod, gre:make([{'a', 1}], [list])),
+                    glc:handle(Mod, gre:make([{'b', 2}], [list])),
+
+                    {compiled, Mod} = setup_query(testmod17c,
+                        glc:any([glc:eq(a, 1), glc:eq(b, 2)]), [], false),
+                    ?assertEqual(4, Mod:info(input)),
+                    ?assertEqual(2, Mod:info(filter)),
+                    ?assertEqual(2, Mod:info(output)),
+
+                    glc:reset_counters(Mod, input),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(2, Mod:info(filter)),
+                    ?assertEqual(2, Mod:info(output)),
+                    glc:reset_counters(Mod, filter),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(2, Mod:info(output)),
+                    glc:reset_counters(Mod),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output))
+                end
+            },
             {"ets data recovery test",
                 fun() ->
                     Self = self(),
@@ -709,16 +865,35 @@ events_test_() ->
                     ?assertEqual(7, length(gr_counter:list(Mod:table(counters))))
                 end
             },
+            {"ets data recovery test no stats",
+                fun() ->
+                    Self = self(),
+                    {compiled, Mod} = setup_query(testmod18b,
+                        glc:with(glc:eq(a, 1), fun(Event) -> Self ! gre:fetch(a, Event) end),
+                            [{statistics, false}]),
+                    glc:handle(Mod, gre:make([{a,1}], [list])),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(1, receive Msg -> Msg after 0 -> notcalled end),
+                    ?assertEqual(1, length(gr_param:list(Mod:table(params)))),
+                    true = exit(whereis(Mod:table(params)), kill),
+                    ?assertEqual(undefined, whereis(Mod:table(counters))),
+                    ?assertEqual(0, Mod:info(input)),
+                    glc:handle(Mod, gre:make([{'a', 1}], [list])),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(1, length(gr_param:list(Mod:table(params))))
+                end
+            },
             {"run timed job test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
-                    Runtime = 0.15,
+                    Store = [{stored, value}, {statistics, true}],
+                    Runtime = 0.015,
                     {compiled, Mod} = setup_query(testmod19,
                                                   glc:gt(runtime, Runtime),
                                                   Store),
                     glc:run(Mod, fun(Event, EStore) -> 
-                        timer:sleep(100),
+                        timer:sleep(10),
                         Self ! {gre:fetch(a, Event), EStore}
                     end, gre:make([{a,1}], [list])),
                     ?assertEqual(0, Mod:info(output)),
@@ -731,7 +906,7 @@ events_test_() ->
                                                   Store),
                     glc:handle(Mod, gre:make([{'a', 1}], [list])),
                     glc:run(Mod, fun(Event, EStore) -> 
-                        timer:sleep(200),
+                        timer:sleep(30),
                         Self ! {gre:fetch(a, Event), EStore}
                     end, gre:make([{a,2}], [list])),
                     ?assertEqual(1, Mod:info(output)),
@@ -742,8 +917,11 @@ events_test_() ->
             },
             {"reset job counters",
                 fun() ->
-                    {compiled, Mod} = setup_query(testmod20,
-                        glc:any([glc:eq(a, 1), glc:gt(runtime, 0.15)])),
+                    Self = self(),
+                    Store = [{stored, value}, {statistics, true}],
+
+                    {compiled, Mod} = setup_query(testmod20a,
+                        glc:any([glc:eq(a, 1), glc:gt(runtime, 0.015)]), Store),
                     glc:handle(Mod, gre:make([{'a', 2}], [list])),
                     glc:handle(Mod, gre:make([{'b', 1}], [list])),
                     ?assertEqual(2, Mod:info(input)),
@@ -754,17 +932,20 @@ events_test_() ->
                     ?assertEqual(3, Mod:info(filter)),
                     ?assertEqual(1, Mod:info(output)),
 
-                    Self = self(),
                     glc:run(Mod, fun(Event, EStore) -> 
-                        timer:sleep(100),
+                        timer:sleep(20),
                         Self ! {gre:fetch(a, Event), EStore}
                     end, gre:make([{a,1}], [list])),
                     ?assertEqual(2, Mod:info(output)),
                     ?assertEqual(3, Mod:info(filter)),
-                    ?assertEqual(1, receive {Msg, undefined} -> Msg after 0 -> notcalled end),
+                    receive {Msg, _} ->
+                        ?assertEqual(1, Msg)
+                    after 0 ->
+                        erlang:error(notcalled)
+                    end,
 
                     {_, Msg1} = glc:run(Mod, fun(_Event, _EStore) -> 
-                        timer:sleep(200),
+                        timer:sleep(20),
                         {error, badtest}
                         
                     end, gre:make([{a,1}], [list])),
@@ -776,7 +957,7 @@ events_test_() ->
                     ?assertEqual({error, badtest}, Msg1),
 
                     {_, Msg2} = glc:run(Mod, fun(_Event, _EStore) -> 
-                        timer:sleep(20),
+                        timer:sleep(10),
                         {ok, goodtest}
                         
                     end, gre:make([{a,2}], [list])),
@@ -832,10 +1013,110 @@ events_test_() ->
                     ?assertEqual(0, Mod:info(job_run))
                 end
             },
+            {"reset job counters with no statistics",
+                fun() ->
+                    Self = self(),
+                    Store = [{stored, value}, {statistics, false}],
+
+                    {compiled, Mod} = setup_query(testmod20b,
+                        glc:any([glc:eq(a, 1), glc:gt(runtime, 0.15)]), Store),
+                    glc:handle(Mod, gre:make([{'a', 2}], [list])),
+                    glc:handle(Mod, gre:make([{'b', 1}], [list])),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    glc:handle(Mod, gre:make([{'a', 1}], [list])),
+                    glc:handle(Mod, gre:make([{'b', 2}], [list])),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output)),
+
+                    glc:run(Mod, fun(Event, EStore) -> 
+                        Self ! {gre:fetch(a, Event), EStore}
+                    end, gre:make([{a,1}], [list])),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(filter)),
+
+                    % not working?
+                    % ?assertEqual(1, receive {Msg, undefined} -> Msg after 0 -> notcalled end),
+                    receive {Msg, _} ->
+                        ?assertEqual(1, Msg)
+                    after 0 ->
+                        erlang:error(notcalled)
+                    end,
+
+                    {_, Msg1} = glc:run(Mod, fun(_Event, _EStore) -> 
+                        timer:sleep(20),
+                        {error, badtest}
+                        
+                    end, gre:make([{a,1}], [list])),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_error)),
+                    ?assertEqual(0, Mod:info(job_run)),
+                    ?assertEqual({error, badtest}, Msg1),
+
+                    {_, Msg2} = glc:run(Mod, fun(_Event, _EStore) -> 
+                        timer:sleep(20),
+                        {ok, goodtest}
+                        
+                    end, gre:make([{a,2}], [list])),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_error)),
+                    ?assertEqual(0, Mod:info(job_run)),
+                    ?assertEqual({ok, goodtest}, Msg2),
+
+
+                    glc:reset_counters(Mod, input),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_error)),
+                    ?assertEqual(0, Mod:info(job_run)),
+                    glc:reset_counters(Mod, filter),
+                    ?assertEqual(0, glc:input(Mod)),
+                    ?assertEqual(0, glc:filter(Mod)),
+                    ?assertEqual(0, glc:output(Mod)),
+                    ?assertEqual(0, glc:job_input(Mod)),
+                    ?assertEqual(0, glc:job_error(Mod)),
+                    ?assertEqual(0, glc:job_run(Mod)),
+                    glc:reset_counters(Mod, output),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_error)),
+                    ?assertEqual(0, Mod:info(job_run)),
+                    glc:reset_counters(Mod, job_input),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_error)),
+                    ?assertEqual(0, Mod:info(job_run)),
+                    glc:reset_counters(Mod, job_error),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_error)),
+                    ?assertEqual(0, Mod:info(job_run)),
+                    glc:reset_counters(Mod, job_run),
+                    ?assertEqual(0, Mod:info(input)),
+                    ?assertEqual(0, Mod:info(filter)),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(0, Mod:info(job_input)),
+                    ?assertEqual(0, Mod:info(job_error)),
+                    ?assertEqual(0, Mod:info(job_run))
+                end
+            },
             {"variable storage test",
                 fun() ->
-                    {compiled, Mod} = setup_query(testmod20a,
-                        glc:eq(a, 2), [{stream, time}]),
+                    {compiled, Mod} = setup_query(testmod20c,
+                        glc:eq(a, 2), [{stream, time}, {statistics, true}]),
                     glc:handle(Mod, gre:make([{'a', 2}], [list])),
                     glc:handle(Mod, gre:make([{'b', 1}], [list])),
                     ?assertEqual(2, Mod:info(input)),
@@ -847,17 +1128,35 @@ events_test_() ->
                     ?assertEqual({error, undefined}, glc:get(Mod, beam))
                 end
             },
-            {"with multi function any test",
+            {"with multi function any test no stats",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, false}],
 
                     G1 = glc:with(glc:eq(a, 1), fun(_Event, EStore) -> 
                        Self ! {a, EStore} end),
                     G2 = glc:with(glc:eq(b, 2), fun(_Event, EStore) -> 
                        Self ! {b, EStore} end),
 
-                    {compiled, Mod} = setup_query(testmod20b, any([G1, G2]),
+                    {compiled, Mod} = setup_query(testmod20d, any([G1, G2]),
+                         Store),
+                    glc:handle(Mod, gre:make([{a,1}], [list])),
+                    ?assertEqual(0, Mod:info(output)),
+                    ?assertEqual(a, receive {Msg, _Store} -> Msg after 0 -> notcalled end),
+                    ?assertEqual(b, receive {Msg, _Store} -> Msg after 0 -> notcalled end)
+                end
+            },
+            {"with multi function any test",
+                fun() ->
+                    Self = self(),
+                    Store = [{stored, value}, {statistics, true}],
+
+                    G1 = glc:with(glc:eq(a, 1), fun(_Event, EStore) -> 
+                       Self ! {a, EStore} end),
+                    G2 = glc:with(glc:eq(b, 2), fun(_Event, EStore) -> 
+                       Self ! {b, EStore} end),
+
+                    {compiled, Mod} = setup_query(testmod20e, any([G1, G2]),
                          Store),
                     glc:handle(Mod, gre:make([{a,1}], [list])),
                     ?assertEqual(1, Mod:info(output)),
@@ -868,7 +1167,7 @@ events_test_() ->
             {"with multi function all test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, true}],
 
                     G1 = glc:with(glc:eq(a, 1), fun(_Event, EStore) -> 
                        Self ! {a, EStore} end),
@@ -895,7 +1194,7 @@ events_test_() ->
             {"with multi-function output match test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, true}],
 
                     {compiled, Mod} = setup_query(testmod22,
                           [glc:with(glc:eq(a, 1), fun(Event, _EStore) -> 
@@ -912,7 +1211,7 @@ events_test_() ->
             {"with multi-function output double-match test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, true}],
                     {compiled, Mod} = setup_query(testmod23,
                           [glc:with(glc:eq(a, 1), fun(Event, _EStore) -> 
                               Self ! {a, gre:fetch(a, Event)} end),
@@ -928,7 +1227,7 @@ events_test_() ->
             {"with multi function complex match test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, true}],
 
                     G1 = glc:with(glc:gt(r, 0.1), fun(_Event, EStore) -> 
                        Self ! {a, EStore} end),
@@ -971,20 +1270,20 @@ events_test_() ->
             {"with single-function run test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, true}],
                     {compiled, Mod1} = setup_query(testmod25a,
-                        glc:with(glc:all([glc:gt(runtime, 0.15), glc:lt(a, 3)]), fun(Event, EStore) -> 
+                        glc:with(glc:all([glc:gt(runtime, 0.015), glc:lt(a, 3)]), fun(Event, EStore) -> 
                            Self ! {gre:fetch(a, Event), EStore} end),
                          Store),
-                    glc:run(Mod1, fun(_Event, _EStore) -> timer:sleep(200), ok end, gre:make([{a, 2}], [list])),
+                    glc:run(Mod1, fun(_Event, _EStore) -> timer:sleep(20), ok end, gre:make([{a, 2}], [list])),
                     ?assertEqual(1, Mod1:info(output)),
                     ?assertEqual(2, receive {Msg, Store} -> Msg after 250 -> notcalled end),
                     {compiled, Mod2} = setup_query(testmod25b,
-                        glc:with(glc:all([glc:gt(runtime, 0.15), glc:lt(a, 3)]), fun(Event, EStore) -> 
+                        glc:with(glc:all([glc:gt(runtime, 0.015), glc:lt(a, 3)]), fun(Event, EStore) -> 
                            Self ! {gre:fetch(a, Event), EStore}
                         end), Store),
                     {_, {error, later}} = glc:run(Mod2, fun(_Event, _EStore) -> 
-                           timer:sleep(200), 
+                           timer:sleep(20), 
                            erlang:exit(later)
                     end, gre:make([{a, 2}], [list])),
                     ?assertEqual(1, Mod2:info(output)),
@@ -995,9 +1294,9 @@ events_test_() ->
             {"with multi-function output run error test",
                 fun() ->
                     Self = self(),
-                    Store = [{stored, value}],
+                    Store = [{stored, value}, {statistics, true}],
                     {compiled, Mod} = setup_query(testmod26,
-                          [glc:with(glc:gt(runtime, 0.15), fun(Event, _EStore) -> 
+                          [glc:with(glc:gt(runtime, 0.015), fun(Event, _EStore) -> 
                               Self ! {a, gre:fetch(b, Event)} 
                            end),
                            glc:with(glc:eq(c, 3), fun(Event, _EStore) -> 
@@ -1015,7 +1314,7 @@ events_test_() ->
                          Store),
                     Event = gre:make([{a,1}, {b, 3}, {c, 4}], [list]),
                     {_, {error, bye}} = glc:run(Mod, fun(_Event, _EStore) -> 
-                            timer:sleep(200),
+                            timer:sleep(20),
                             erlang:error(bye) 
                     end, Event),
                     ?assertEqual(3, Mod:info(output)),
@@ -1023,6 +1322,22 @@ events_test_() ->
                     ?assertEqual(1, Mod:info(job_error)),
                     ?assertEqual(b, receive {b=Msg, _Store} -> Msg after 0 -> notcalled end)
                 end
+            },
+            {"with pid storage test",
+                fun() ->
+                    Self = self(),
+                    XPid = spawn(fun() -> receive {msg, Msg, Pid} -> Self ! {Msg, Pid} end end),
+                    Store = [{stored, XPid}, {statistics, true}],
+                    {compiled, Mod} = setup_query(testmod27,
+                        glc:with(glc:eq(a, 1), fun(Event, _EStore) -> 
+                           {ok, Pid} = glc:get(testmod27, stored),
+                           Pid ! {msg, gre:fetch(a, Event), Self}
+                         end),
+                         Store),
+                    glc:handle(Mod, gre:make([{a,1}], [list])),
+                    ?assertEqual(1, Mod:info(output)),
+                    ?assertEqual(1, receive {Msg, Pid} -> Pid ! Msg after 2 -> notcalled end)
+                end
             }
         ]
     }.
